using System;
using System.IO;
using System.Text;
using System.Text.RegularExpressions;
using System.Threading;
using System.Xml;

namespace MapData
{
	/// <summary>
	/// The purpose of this class is to map a flat file to an XML document and to map an 
	/// XML document into a flat file.  Mapping can be performed on both comma delimited 
	/// and fixed-field length flat files.
	/// </summary>
	/// <remarks>
	/// Class:				DataMapper
	/// Programmer:			Edmond Woychowsky
	/// Creation Date:		September 20, 2004
	/// Last Modification:	
	/// </remarks>
	public class DataMapper 
	{
		#region static
		private static string COMMA_DELIMITED = "comma_delimited";
		private static string CRLF = "\r\n";		//	Carriage return/line feed
		private static string FIELD_LENGTH = "field_length";
		private static string FIELD_NAME = "field_name";
		private static string FIELD_OFFSET = "field_offset";
		private static string FIELD_TYPE = "field_type";
		private static string KEY_CONTENT = "key_content";
		private static string KEY_LENGTH = "key_length";
		private static string KEY_OFFSET = "key_offset";
		private static string KEY_NAME = "key_name";
		private static string KEY_TYPE = "key_type";
		private static string NUMERIC = "n";		//	Numeric field
		private static string RECORD_NAME = "record_name";
		private static string ROOT = "file";
		private static string XPATH_RECORDS = "//record";
		private static string XPATH_FIELDS = "//field";
		private static string DELIMITER = ",";		//	Comma field delimiter
		private static char NEWLINE = '\n';			//	Line feed
		#endregion

		#region private
		private XmlDocument m_map_xml = new XmlDocument();
		private XmlDocument m_xml = new XmlDocument();
		private Dictionary m_records = new Dictionary();
		private StreamReader m_input_stream;		//	Input stream
		private int m_active_threads = 0;			//	Active thread count
		#endregion

		#region constructors
		/*
		 * Method:			DataMapper (overloaded)
		 * Programmer:		Edmond Woychowsky
		 * Creation Date:	September 20, 2004
		 * Description:		
		 */
		/// <summary>
		/// Class Constructor.
		/// </summary>
		public DataMapper() { }

		/// <summary>
		/// Class Constructor.
		/// </summary>
		/// <param name="map">Name of the file containing the XML data map</param>
		public DataMapper(string map) 
		{
			Map(map);
		}

		/// <summary>
		/// Class Constructor.
		/// </summary>
		/// <param name="map">An XML document containing the data map</param>
		public DataMapper(XmlDocument map) 
		{
			Map(map);
		}

		/// <summary>
		/// Class Constructor.
		/// </summary>
		/// <param name="map">Name of the file containing the XML data map</param>
		/// <param name="input">Name of the file containg the data to be mapped</param>
		public DataMapper(string map,string input) 
		{
			Map(map);
			Input(input);
		}

		/// <summary>
		/// Class Constructor.
		/// </summary>
		/// <param name="map">Name of the file containing the XML data map</param>
		/// <param name="input">An XML document containing the data to be mapped</param>
		public DataMapper(string map,XmlDocument input) 
		{
			Map(map);
			Input(input);
		}

		/// <summary>
		/// Class Constructor.
		/// </summary>
		/// <param name="map">Name of the file containing the XML data map</param>
		/// <param name="input">A stream containing the data to be mapped</param>
		public DataMapper(string map,Stream input) 
		{
			Map(map);
			Input(input);
		}

		/// <summary>
		/// Class Constructor.
		/// </summary>
		/// <param name="map">Name of the file containing the XML data map</param>
		/// <param name="input">A stream containing the data to be mapped</param>
		public DataMapper(string map,StreamReader input) 
		{
			Map(map);
			Input(input);
		}

		/// <summary>
		/// Class Constructor.
		/// </summary>
		/// <param name="map">Name of the file containing the XML data map</param>
		/// <param name="input">A stream containing the data to be mapped</param>
		public DataMapper(string map,MemoryStream input) 
		{
			Map(map);
			Input(input);
		}

		/// <summary>
		/// Class Constructor.
		/// </summary>
		/// <param name="map">An XML document containing the data map</param>
		/// <param name="input">Name of the file containg the data to be mapped</param>
		public DataMapper(XmlDocument map,string input) 
		{
			Map(map);
			Input(input);
		}

		/// <summary>
		/// Class Constructor.
		/// </summary>
		/// <param name="map">An XML document containing the data map</param>
		/// <param name="input">An XML document containing the data to be mapped</param>
		public DataMapper(XmlDocument map,XmlDocument input) 
		{
			Map(map);
			Input(input);
		}

		/// <summary>
		/// Class Constructor.
		/// </summary>
		/// <param name="map">An XML document containing the data map</param>
		/// <param name="input">A stream containing the data to be mapped</param>
		public DataMapper(XmlDocument map,Stream input) 
		{
			Map(map);
			Input(input);
		}

		/// <summary>
		/// Class Constructor.
		/// </summary>
		/// <param name="map">An XML document containing the data map</param>
		/// <param name="input">A stream containing the data to be mapped</param>
		public DataMapper(XmlDocument map,StreamReader input) 
		{
			Map(map);
			Input(input);
		}

		/// <summary>
		/// Class Constructor.
		/// </summary>
		/// <param name="map">An XML document containing the data map</param>
		/// <param name="input">A stream containing the data to be mapped</param>
		public DataMapper(XmlDocument map,MemoryStream input) 
		{
			Map(map);
			Input(input);
		}
		#endregion

		#region publicMethods
		/// <summary>
		/// Wrap up processing.
		/// </summary>
		public void Close() 
		{
			threadWait();							//	Wait for outstanding threads
		}

		/// <summary>
		/// Map the provided input to XML.
		/// </summary>
		/// <returns>The mapped input</returns>
		public Stream Data() 
		{
			threadWait();							//	Wait for outstanding threads

			MemoryStream stream = new MemoryStream();

			for(int i=0;i < this.m_xml.FirstChild.ChildNodes.Count;i++) 
			{
				XmlNode record = this.m_xml.FirstChild.ChildNodes.Item(i);
				Dictionary dictionary = (Dictionary)this.m_records.Item(record.Name);
				StringBuilder line = new StringBuilder();
				StringBuilder xpath = new StringBuilder();

				xpath.Append(XPATH_RECORDS);
				xpath.Append("[@");
				xpath.Append(RECORD_NAME);
				xpath.Append(" = '");
				xpath.Append((string)dictionary.Item(RECORD_NAME));
				xpath.Append("']");

				XmlNode map = this.m_map_xml.SelectSingleNode(xpath.ToString());

				for(int j=0;j < record.ChildNodes.Count;j++)
				{
					StringBuilder innerText = new StringBuilder();

					if(bool.Parse((string)dictionary.Item(COMMA_DELIMITED))) 
						if(j != 0)
							line.Append(DELIMITER);

					xpath = new StringBuilder();	//	Build XPath

					xpath.Append(XPATH_FIELDS);
					xpath.Append("[@");
					xpath.Append(FIELD_NAME);
					xpath.Append(" = '");
					xpath.Append((string)record.ChildNodes.Item(j).Name);
					xpath.Append("']");

					try 
					{
						XmlNode field = map.SelectSingleNode(xpath.ToString());
						int length = int.Parse(field.Attributes.GetNamedItem(FIELD_LENGTH).InnerText.ToString());

						if(bool.Parse((string)dictionary.Item(COMMA_DELIMITED))) 
							innerText.Append(record.ChildNodes.Item(j).InnerText.ToString());
						else
							if(field.Attributes.GetNamedItem(FIELD_TYPE).InnerText.ToString() == NUMERIC) 
						{
							if(record.ChildNodes.Item(j).InnerText.ToString().Length < length)
								for(int k = record.ChildNodes.Item(j).InnerText.ToString().Length;k < length;k++)
									innerText.Append("0");

							innerText.Append(record.ChildNodes.Item(j).InnerText.ToString());
						} 
						else 
						{
							innerText.Append(record.ChildNodes.Item(j).InnerText.ToString());

							if(record.ChildNodes.Item(j).InnerText.ToString().Length < length)
								for(int k = record.ChildNodes.Item(j).InnerText.ToString().Length;k < length;k++)
									innerText.Append(" ");
						}
					}
					catch(Exception) 
					{
						innerText.Append(record.ChildNodes.Item(j).InnerText.ToString());
					}
					finally 
					{
						line.Append(innerText.ToString());
					}
				}

				line.Append(NEWLINE);

				byte[] buffer = Encoding.ASCII.GetBytes(line.ToString());

				stream.Write(buffer,0,buffer.Length);
			}

			byte[] eof = Encoding.ASCII.GetBytes(CRLF);

			stream.Write(eof,0,eof.Length);

			stream.Position = 0;					//	Position at beginning

			return(stream);
		}

		/// <summary>
		/// Fork a thread that loads the input to be mapped.
		/// </summary>
		/// <param name="path">Full path of input file</param>
		public void Input(string path) 
		{
			Interlocked.Increment(ref this.m_active_threads);

			ThreadPool.QueueUserWorkItem(new WaitCallback(loadInput),path);
		}

		/// <summary>
		/// Fork a thread that loads the input to be mapped.
		/// </summary>
		/// <param name="xml">Input to be mapped</param>
		public void Input(XmlDocument xml) 
		{
			Interlocked.Increment(ref this.m_active_threads);

			ThreadPool.QueueUserWorkItem(new WaitCallback(loadInput),xml);
		}

		/// <summary>
		/// Fork a thread that loads the input to be mapped.
		/// </summary>
		/// <param name="stream">Input to be mapped</param>
		public void Input(Stream stream) 
		{
			Interlocked.Increment(ref this.m_active_threads);

			ThreadPool.QueueUserWorkItem(new WaitCallback(loadInput),stream);
		}

		/// <summary>
		/// Fork a thread that loads the input to be mapped.
		/// </summary>
		/// <param name="stream">Input to be mapped</param>
		public void Input(StreamReader stream) 
		{
			Interlocked.Increment(ref this.m_active_threads);

			ThreadPool.QueueUserWorkItem(new WaitCallback(loadInput),stream);
		}

		/// <summary>
		/// Fork a thread that loads the input to be mapped.
		/// </summary>
		/// <param name="stream">Input to be mapped</param>
		public void Input(MemoryStream stream) 
		{
			Interlocked.Increment(ref this.m_active_threads);

			ThreadPool.QueueUserWorkItem(new WaitCallback(loadInput),stream);
		}

		/// <summary>
		/// Load the XML document that defines the data mapping.  In addition, some 
		/// pre-processing of the map is performed.
		/// </summary>
		/// <param name="state">Data map</param>
		private void loadMap(object state) 
		{
			lock(this.m_map_xml) 
			{
				switch(state.GetType().ToString()) 
				{
					case "System.String":
						this.m_map_xml.Load((string) state);

						break;
					case "System.Xml.XmlDocument":
						MemoryStream stream = new MemoryStream();

						((XmlDocument)state).Save(stream);

						stream.Position = 0;		//	Position at start

						this.m_map_xml.Load(stream);

						break;
				}
			}

			for(int i=0;i < this.m_map_xml.SelectNodes(XPATH_RECORDS).Count;i++) 
			{
				Dictionary record = parseNode(this.m_map_xml.SelectNodes(XPATH_RECORDS).Item(i));

				if(record.Contains(KEY_NAME)) 
				{
					StringBuilder xpath = new StringBuilder();

					xpath.Append(XPATH_FIELDS);
					xpath.Append("[@");
					xpath.Append(FIELD_NAME);
					xpath.Append(" = '");
					xpath.Append(record.Item(KEY_NAME));
					xpath.Append("']");

					Dictionary field = parseNode(this.m_map_xml.SelectNodes(XPATH_RECORDS).Item(i).SelectSingleNode(xpath.ToString()));

					record.Add(KEY_LENGTH,field.Item(FIELD_LENGTH));
					record.Add(KEY_OFFSET,field.Item(FIELD_OFFSET));
					record.Add(KEY_TYPE,field.Item(FIELD_TYPE));
				}

				this.m_records.Add(record.Item(RECORD_NAME),record);
			}

			Interlocked.Decrement(ref this.m_active_threads);
		}

		/// <summary>
		/// Fork a thread that loads the XML document that defines the data mapping.
		/// </summary>
		/// <param name="path">Full path of map file</param>
		public void Map(string path) 
		{
			Interlocked.Increment(ref this.m_active_threads);

			ThreadPool.QueueUserWorkItem(new WaitCallback(loadMap),path);
		}

		/// <summary>
		/// Fork a thread that loads the XML document that defines the data mapping.
		/// </summary>
		/// <param name="map">Data map</param>
		public void Map(XmlDocument map) 
		{
			Interlocked.Increment(ref this.m_active_threads);

			ThreadPool.QueueUserWorkItem(new WaitCallback(loadMap),map);
		}

		/// <summary>
		/// Map the provided input to XML.
		/// </summary>
		/// <returns>Mapped input</returns>
		public Stream Xml() 
		{
			threadWait();							//	Wait for outstanding threads

			Stream stream = new MemoryStream();
			XmlTextWriter writer = new XmlTextWriter(stream,null);
			Dictionary dictionary = null;			//	Initialize dictionary
			string line = this.m_input_stream.ReadLine();

			writer.WriteStartElement(ROOT);

			do 
			{
				dictionary = (Dictionary)this.m_records.Item(recordMap(line));

				try 
				{
					writer.WriteStartElement((string)dictionary.Item(RECORD_NAME));
				}
				catch(Exception) { }

				StringBuilder xpath = new StringBuilder();

				xpath.Append(XPATH_RECORDS);	//	Build record XPath
				xpath.Append("[@");
				xpath.Append(RECORD_NAME);
				xpath.Append(" = '");
				xpath.Append((string)dictionary.Item(RECORD_NAME));
				xpath.Append("']");

				XmlNode map = this.m_map_xml.SelectSingleNode(xpath.ToString());

				if(bool.Parse((string)dictionary.Item(COMMA_DELIMITED)))
				{
					Regex re = new Regex(DELIMITER);
					string[] fields = re.Split(line);
					string text = null;

					for(int i=0;i < map.ChildNodes.Count;i++) 
					{
						try 
						{
							if(fields[i].Length != 0)
								text = fields[i];
							else
								text = null;		//	Default null
						}
						catch(Exception) 
						{
							text = null;
						}
						finally 
						{
							writer.WriteElementString(map.ChildNodes.Item(i).Attributes.GetNamedItem(FIELD_NAME).InnerText.ToString(),text);
						}
					}
				} 
				else
				{
					for(int i=0;i < map.ChildNodes.Count;i++) 
					{
						string text = null;

						try 
						{
							int offset = int.Parse(map.ChildNodes.Item(i).Attributes.GetNamedItem(FIELD_OFFSET).InnerText.ToString());
							int length = int.Parse(map.ChildNodes.Item(i).Attributes.GetNamedItem(FIELD_LENGTH).InnerText.ToString());

							text = line.Substring(offset,length);
						}
						catch(Exception) { }
						finally 
						{
							writer.WriteElementString(map.ChildNodes.Item(i).Attributes.GetNamedItem(FIELD_NAME).InnerText.ToString(),text);
						}
					}
				}

				writer.WriteEndElement();

				line = this.m_input_stream.ReadLine();
			} while(line != null);

			writer.WriteEndElement();
			writer.Flush();

			stream.Position = 0;

			return(stream);
		}
		#endregion

		#region privateMethods
		/// <summary>
		/// Load the input to be mapped.
		/// </summary>
		/// <param name="state">Input to be mapped</param>
		private void loadInput(object state) 
		{
			lock(this.m_xml) 
			{
				this.m_xml = new XmlDocument();

				switch(state.GetType().ToString()) 
				{
					case "System.String":
						try 
						{
							lock(this.m_xml) 
							{
								this.m_xml.Load((string)state);
							}
						}
						catch(Exception) 
						{
							this.m_input_stream = new StreamReader((string)state);
						}

						break;
					case "System.Xml.XmlDocument":
						MemoryStream stream = new MemoryStream();

						((XmlDocument)state).Save(stream);

						stream.Position = 0;		//	Position at start

						this.m_xml.Load(stream);

						break;
					case "System.IO.StreamReader":
						this.m_input_stream = (StreamReader)state;

						break;
					case "System.IO.MemoryStream":
						this.m_input_stream = new StreamReader((MemoryStream)state);

						break;
				}
			}

			Interlocked.Decrement(ref this.m_active_threads);
		}

		/// <summary>
		/// Build a dictionary collection consisting of attributes of the passed node.
		/// </summary>
		/// <param name="node">Current node to process</param>
		/// <returns>Dictionary consisting of the node's attributes</returns>
		private Dictionary parseNode(XmlNode node) 
		{
			XmlAttributeCollection attribute = node.Attributes;
			Dictionary record = new Dictionary();

			for(int i=0;i < attribute.Count;i++) 
			{
				record.Add(attribute.Item(i).Name,attribute.Item(i).InnerText.ToString());
			}

			return(record);
		}

		/// <summary>
		/// Determine the recordname for the current record.
		/// </summary>
		/// <param name="record">Current record</param>
		/// <returns>Current record name</returns>
		private string recordMap(string record) 
		{
			Dictionary dictionary = null;
			object[] keys = this.m_records.Keys();
			string recordName = null;

			for(int i=0;i < keys.Length;i++) 
			{
				dictionary = (Dictionary)this.m_records.Item(keys[i]);

				if(dictionary.Contains(KEY_NAME)) 
				{
					string contents = null;

					try 
					{
						contents = record.Substring(int.Parse((string)dictionary.Item(KEY_OFFSET)),int.Parse((string)dictionary.Item(KEY_LENGTH)));
					}
					catch(Exception) 
					{
						if(dictionary.Contains(COMMA_DELIMITED))
							if(bool.Parse(dictionary.Item(COMMA_DELIMITED).ToString())) 
							{
								StringBuilder xpath = new StringBuilder();

								xpath.Append(XPATH_RECORDS);	//	Build record XPath
								xpath.Append("[@");
								xpath.Append(RECORD_NAME);
								xpath.Append(" = '");
								xpath.Append((string)dictionary.Item(RECORD_NAME));
								xpath.Append("']");

								XmlNode map = this.m_map_xml.SelectSingleNode(xpath.ToString());
								Regex re = new Regex(DELIMITER);
								string[] fields = re.Split(record);

								for(int j=0;j < fields.Length;j++) 
									if(map.ChildNodes.Item(j).Attributes.GetNamedItem(FIELD_NAME).InnerText.ToString() == dictionary.Item(KEY_NAME).ToString())
									{
										contents = fields[j];

										break;
									}
							}
					}

					if(contents == dictionary.Item(KEY_CONTENT).ToString()) 
					{
						recordName = (string)dictionary.Item(RECORD_NAME);

						break;
					}
				}
				else
					recordName = (string)dictionary.Item(RECORD_NAME);
			}

			return(recordName);
		}

		/// <summary>
		/// Wait for all outstanding threads to terminate.
		/// </summary>
		private void threadWait() 
		{
			while(Interlocked.CompareExchange(ref this.m_active_threads,0,0) != 0)
				Thread.Sleep(10);
		}
		#endregion
	}
}
